@@ -13,34 +13,31 @@
 /*	} BOUNCE_INFO;
 /*
 /*	BOUNCE_INFO *bounce_mail_init(service, queue_name, queue_id,
-/*					encoding, dsn_envid, template)
+/*					encoding, dsn_envid, report_type)
 /*	const char *service;
 /*	const char *queue_name;
 /*	const char *queue_id;
 /*	const char *encoding;
 /*	const char *dsn_envid;
-/*	const BOUNCE_TEMPLATE *template;
+/*	int	report_type;
 /*
 /*	BOUNCE_INFO *bounce_mail_one_init(queue_name, queue_id, encoding,
-/*					dsn_envid, dsn_notify, rcpt_buf,
-/*					dsn_buf, template)
+/*					dsn_envid, dsn_notify, rcpt, dsn)
 /*	const char *queue_name;
 /*	const char *queue_id;
 /*	const char *encoding;
 /*	int	dsn_notify;
 /*	const char *dsn_envid;
-/*	RCPT_BUF *rcpt_buf;
-/*	DSN_BUF	*dsn_buf;
-/*	const BOUNCE_TEMPLATE *template;
+/*	RECIPIENT *rcpt;
+/*	DSN	*dsn;
 /*
 /*	void	bounce_mail_free(bounce_info)
 /*	BOUNCE_INFO *bounce_info;
 /*
-/*	int	bounce_header(fp, bounce_info, recipient, postmaster_copy)
+/*	int	bounce_header(fp, bounce_info, recipient)
 /*	VSTREAM *fp;
 /*	BOUNCE_INFO *bounce_info;
 /*	const char *recipient;
-/*	int	postmaster_copy;
 /*
 /*	int	bounce_boilerplate(fp, bounce_info)
 /*	VSTREAM *fp;
@@ -86,18 +83,21 @@
 /*	bounce_mail_init() bundles up its argument and attempts to
 /*	open the corresponding logfile and message file. A BOUNCE_INFO
 /*	structure contains all the necessary information about an
-/*	undeliverable message.
+/*	undeliverable message. The report type is BOUNCE_REPORT_WARN
+/*	for delayed mail, BOUNCE_REPORT_FAIL for undeliverable mail,
+/*	BOUNCE_REPORT_SUCCESS for "success" delivery notification,
+/*	or BOUNCE_REPORT_OTHER for other status reports.
 /*
 /*	bounce_mail_one_init() provides the same function for only
-/*	one recipient that is not read from bounce logfile.
+/*	one recipient that is not read from bounce logfile. It
+/*	assumes a report type of BOUNCE_REPORT_FAIL.
 /*
 /*	bounce_mail_free() releases memory allocated by bounce_mail_init()
 /*	and closes any files opened by bounce_mail_init().
 /*
 /*	bounce_header() produces a standard mail header with the specified
 /*	recipient and starts a text/plain message segment for the
-/*	human-readable problem description. postmaster_copy is either
-/*	POSTMASTER_COPY or NO_POSTMASTER_COPY.
+/*	human-readable problem description.
 /*
 /*	bounce_boilerplate() produces the standard "sorry" text that
 /*	creates the illusion that mail systems are civilized.
@@ -160,7 +160,6 @@
 #include <sys_defs.h>
 #include <sys/stat.h>
 #include <stdlib.h>
-#include <stdio.h>			/* sscanf() */
 #include <unistd.h>
 #include <errno.h>
 #include <string.h>
@@ -212,16 +211,14 @@ static BOUNCE_INFO *bounce_mail_alloc(const char *service,
 				              const char *queue_id,
 				              const char *encoding,
 				              const char *dsn_envid,
-				              RCPT_BUF *rcpt_buf,
-				              DSN_BUF *dsn_buf,
-				              BOUNCE_TEMPLATE *template,
+				              int report_type,
 				              BOUNCE_LOG *log_handle)
 {
     BOUNCE_INFO *bounce_info;
     int     rec_type;
 
     /*
-     * Bundle up a bunch of parameters and initialize information that will
+     * Bundle up a bunch of parameters and initialize information. that will
      * be discovered on the fly.
      */
     bounce_info = (BOUNCE_INFO *) mymalloc(sizeof(*bounce_info));
@@ -242,14 +239,12 @@ static BOUNCE_INFO *bounce_mail_alloc(const char *service,
 	bounce_info->dsn_envid = dsn_envid;
     else
 	bounce_info->dsn_envid = 0;
-    bounce_info->template = template;
+    bounce_info->report_type = report_type;
     bounce_info->buf = vstring_alloc(100);
     bounce_info->sender = vstring_alloc(100);
     bounce_info->arrival_time = 0;
     bounce_info->orig_offs = 0;
     bounce_info->message_size = 0;
-    bounce_info->rcpt_buf = rcpt_buf;
-    bounce_info->dsn_buf = dsn_buf;
     bounce_info->log_handle = log_handle;
 
     /*
@@ -283,9 +278,9 @@ static BOUNCE_INFO *bounce_mail_alloc(const char *service,
 	msg_fatal("open %s %s: %m", service, queue_id);
 
     /*
-     * Get time/size/sender information from the original message envelope
-     * records. If the envelope is corrupted just send whatever we can
-     * (remember this is a best effort, it does not have to be perfect).
+     * Skip over the original message envelope records. If the envelope is
+     * corrupted just send whatever we can (remember this is a best effort,
+     * it does not have to be perfect).
      * 
      * Lock the file for shared use, so that queue manager leaves it alone after
      * restarting.
@@ -299,44 +294,20 @@ static BOUNCE_INFO *bounce_mail_alloc(const char *service,
 		      VSTREAM_PATH(bounce_info->orig_fp));
 	while ((rec_type = rec_get(bounce_info->orig_fp,
 				   bounce_info->buf, 0)) > 0) {
-
-	    /*
-	     * Postfix version dependent: data offset in SIZE record.
-	     */
 	    if (rec_type == REC_TYPE_SIZE) {
-		if (bounce_info->message_size == 0)
-		    sscanf(STR(bounce_info->buf), "%ld %ld",
-			   &bounce_info->message_size,
-			   &bounce_info->orig_offs);
-		if (bounce_info->message_size < 0)
+		if (bounce_info->message_size == 0
+		    && (bounce_info->message_size = atol(STR(bounce_info->buf))) < 0)
 		    bounce_info->message_size = 0;
-		if (bounce_info->orig_offs < 0)
-		    bounce_info->orig_offs = 0;
-	    }
-
-	    /*
-	     * Information for the Arrival-Date: attribute.
-	     */
-	    else if (rec_type == REC_TYPE_TIME) {
+	    } else if (rec_type == REC_TYPE_TIME) {
 		if (bounce_info->arrival_time == 0
 		    && (bounce_info->arrival_time = atol(STR(bounce_info->buf))) < 0)
 		    bounce_info->arrival_time = 0;
-	    }
-
-	    /*
-	     * Information for the X-Postfix-Sender: attribute.
-	     */
-	    else if (rec_type == REC_TYPE_FROM) {
+	    } else if (rec_type == REC_TYPE_FROM) {
 		quote_822_local_flags(bounce_info->sender,
 				      VSTRING_LEN(bounce_info->buf) ?
 				      STR(bounce_info->buf) :
 				      mail_addr_mail_daemon(), 0);
-	    }
-
-	    /*
-	     * Backwards compatibility: no data offset in SIZE record.
-	     */
-	    else if (rec_type == REC_TYPE_MESG) {
+	    } else if (rec_type == REC_TYPE_MESG) {
 		/* XXX Future: sender+recipient after message content. */
 		if (VSTRING_LEN(bounce_info->sender) == 0)
 		    msg_warn("%s: no sender before message content record",
@@ -344,10 +315,6 @@ static BOUNCE_INFO *bounce_mail_alloc(const char *service,
 		bounce_info->orig_offs = vstream_ftell(bounce_info->orig_fp);
 		break;
 	    }
-	    if (bounce_info->orig_offs > 0
-		&& bounce_info->arrival_time > 0
-		&& VSTRING_LEN(bounce_info->sender) > 0)
-		break;
 	}
     }
     return (bounce_info);
@@ -360,12 +327,10 @@ BOUNCE_INFO *bounce_mail_init(const char *service,
 			              const char *queue_id,
 			              const char *encoding,
 			              const char *dsn_envid,
-			              BOUNCE_TEMPLATE *template)
+			              int report_type)
 {
     BOUNCE_INFO *bounce_info;
     BOUNCE_LOG *log_handle;
-    RCPT_BUF *rcpt_buf;
-    DSN_BUF *dsn_buf;
 
     /*
      * Initialize the bounce_info structure. If the bounce log cannot be
@@ -375,18 +340,11 @@ BOUNCE_INFO *bounce_mail_init(const char *service,
      * job. But if the system IS running out of resources, raise a fatal
      * run-time error and force a backoff.
      */
-    if ((log_handle = bounce_log_open(service, queue_id, O_RDONLY, 0)) == 0) {
-	if (errno != ENOENT)
-	    msg_fatal("open %s %s: %m", service, queue_id);
-	rcpt_buf = 0;
-	dsn_buf = 0;
-    } else {
-	rcpt_buf = rcpb_create();
-	dsn_buf = dsb_create();
-    }
+    if ((log_handle = bounce_log_open(service, queue_id, O_RDONLY, 0)) == 0
+	&& errno != ENOENT)
+	msg_fatal("open %s %s: %m", service, queue_id);
     bounce_info = bounce_mail_alloc(service, queue_name, queue_id, encoding,
-				    dsn_envid, rcpt_buf, dsn_buf,
-				    template, log_handle);
+				    dsn_envid, report_type, log_handle);
     return (bounce_info);
 }
 
@@ -396,18 +354,19 @@ BOUNCE_INFO *bounce_mail_one_init(const char *queue_name,
 				          const char *queue_id,
 				          const char *encoding,
 				          const char *dsn_envid,
-				          RCPT_BUF *rcpt_buf,
-				          DSN_BUF *dsn_buf,
-				          BOUNCE_TEMPLATE *template)
+				          RECIPIENT *rcpt,
+				          DSN *dsn)
 {
     BOUNCE_INFO *bounce_info;
+    BOUNCE_LOG *log_handle;
 
     /*
-     * Initialize the bounce_info structure for just one recipient.
+     * Initialize the bounce_info structure. Forge a logfile record for just
+     * one recipient.
      */
+    log_handle = bounce_log_forge(rcpt, dsn);
     bounce_info = bounce_mail_alloc("none", queue_name, queue_id, encoding,
-				    dsn_envid, rcpt_buf, dsn_buf, template,
-				    (BOUNCE_LOG *) 0);
+				 dsn_envid, BOUNCE_REPORT_FAIL, log_handle);
     return (bounce_info);
 }
 
@@ -415,13 +374,9 @@ BOUNCE_INFO *bounce_mail_one_init(const char *queue_name,
 
 void    bounce_mail_free(BOUNCE_INFO *bounce_info)
 {
-    if (bounce_info->log_handle) {
-	if (bounce_log_close(bounce_info->log_handle))
-	    msg_warn("%s: read bounce log %s: %m",
-		     bounce_info->queue_id, bounce_info->queue_id);
-	rcpb_free(bounce_info->rcpt_buf);
-	dsb_free(bounce_info->dsn_buf);
-    }
+    if (bounce_info->log_handle && bounce_log_close(bounce_info->log_handle))
+	msg_warn("%s: read bounce log %s: %m",
+		 bounce_info->queue_id, bounce_info->queue_id);
     if (bounce_info->orig_fp && vstream_fclose(bounce_info->orig_fp))
 	msg_warn("%s: read message file %s %s: %m",
 		 bounce_info->queue_id, bounce_info->queue_name,
@@ -436,9 +391,8 @@ void    bounce_mail_free(BOUNCE_INFO *bounce_info)
 /* bounce_header - generate bounce message header */
 
 int     bounce_header(VSTREAM *bounce, BOUNCE_INFO *bounce_info,
-		              const char *dest, int postmaster_copy)
+		              const char *dest)
 {
-    BOUNCE_TEMPLATE *template = bounce_info->template;
 
     /*
      * Print a minimal bounce header. The cleanup service will add other
@@ -446,22 +400,49 @@ int     bounce_header(VSTREAM *bounce, BOUNCE_INFO *bounce_info,
      */
 #define STREQ(a, b) (strcasecmp((a), (b)) == 0)
 
-    /*
-     * Generic headers.
-     */
-    bounce_template_headers(post_mail_fprintf, bounce, template,
-			    STR(quote_822_local(bounce_info->buf, dest)),
-			    postmaster_copy);
+    post_mail_fprintf(bounce, "From: %s (Mail Delivery System)",
+		      MAIL_ADDR_MAIL_DAEMON);
 
     /*
-     * Auto-Submitted header, as per RFC 3834.
+     * Non-delivery subject line.
      */
-    post_mail_fprintf(bounce, "Auto-Submitted: %s", postmaster_copy ?
-		      "auto-generated" : "auto-replied");
+    if (bounce_info->report_type == BOUNCE_REPORT_FAIL) {
+	post_mail_fputs(bounce, dest == var_bounce_rcpt
+		     || dest == var_2bounce_rcpt || dest == var_delay_rcpt ?
+			"Subject: Postmaster Copy: Undelivered Mail" :
+			"Subject: Undelivered Mail Returned to Sender");
+    }
 
     /*
-     * MIME header. Use 8bit encoding when either the bounced message or the
-     * template requires it.
+     * Delayed mail subject line.
+     */
+    else if (bounce_info->report_type == BOUNCE_REPORT_WARN) {
+	post_mail_fputs(bounce, dest == var_bounce_rcpt
+		     || dest == var_2bounce_rcpt || dest == var_delay_rcpt ?
+			"Subject: Postmaster Warning: Delayed Mail" :
+			"Subject: Delayed Mail (still being retried)");
+    }
+
+    /*
+     * DSN SUCCESS report.
+     */
+    else if (bounce_info->report_type == BOUNCE_REPORT_SUCCESS) {
+	post_mail_fputs(bounce,
+			"Subject: Successful Mail Delivery Report");
+    }
+
+    /*
+     * Address verification report, verbose delivery report.
+     */
+    else {
+	post_mail_fputs(bounce, "Subject: Mail Delivery Status Report");
+    }
+
+    post_mail_fprintf(bounce, "To: %s",
+		      STR(quote_822_local(bounce_info->buf, dest)));
+
+    /*
+     * MIME header.
      */
     post_mail_fprintf(bounce, "MIME-Version: 1.0");
     post_mail_fprintf(bounce, "Content-Type: %s; report-type=%s;",
@@ -469,8 +450,6 @@ int     bounce_header(VSTREAM *bounce, BOUNCE_INFO *bounce_info,
     post_mail_fprintf(bounce, "\tboundary=\"%s\"", bounce_info->mime_boundary);
     if (bounce_info->mime_encoding)
 	post_mail_fprintf(bounce, "Content-Transfer-Encoding: %s",
-		     STREQ(bounce_info->mime_encoding, MAIL_ATTR_ENC_7BIT) ?
-			  bounce_template_encoding(template) :
 			  bounce_info->mime_encoding);
     post_mail_fputs(bounce, "");
     post_mail_fputs(bounce, "This is a MIME-encapsulated message.");
@@ -481,8 +460,7 @@ int     bounce_header(VSTREAM *bounce, BOUNCE_INFO *bounce_info,
     post_mail_fputs(bounce, "");
     post_mail_fprintf(bounce, "--%s", bounce_info->mime_boundary);
     post_mail_fprintf(bounce, "Content-Description: %s", "Notification");
-    post_mail_fprintf(bounce, "Content-Type: %s; charset=%s",
-		      "text/plain", bounce_template_charset(template));
+    post_mail_fprintf(bounce, "Content-Type: %s", "text/plain");
     post_mail_fputs(bounce, "");
 
     return (vstream_ferror(bounce));
@@ -494,9 +472,64 @@ int     bounce_boilerplate(VSTREAM *bounce, BOUNCE_INFO *bounce_info)
 {
 
     /*
-     * Print the boiler-plate text.
+     * Print the message body with the problem report. XXX For now, we use a
+     * fixed bounce template. We could use a site-specific parametrized
+     * template with ${name} macros and we could do wonderful things such as
+     * word wrapping to make the text look nicer. No matter how hard we would
+     * try, receiving bounced mail will always suck.
      */
-    bounce_template_expand(post_mail_fputs, bounce, bounce_info->template);
+#define UNDELIVERED(type) \
+	((type) == BOUNCE_REPORT_FAIL || (type) == BOUNCE_REPORT_WARN)
+
+    post_mail_fprintf(bounce, "This is the %s program at host %s.",
+		      var_mail_name, var_myhostname);
+    post_mail_fputs(bounce, "");
+    if (bounce_info->report_type == BOUNCE_REPORT_FAIL) {
+	post_mail_fputs(bounce,
+	     "I'm sorry to have to inform you that your message could not");
+	post_mail_fputs(bounce,
+	    "be delivered to one or more recipients. It's attached below.");
+    } else if (bounce_info->report_type == BOUNCE_REPORT_WARN) {
+	post_mail_fputs(bounce,
+			"####################################################################");
+	post_mail_fputs(bounce,
+			"# THIS IS A WARNING ONLY.  YOU DO NOT NEED TO RESEND YOUR MESSAGE. #");
+	post_mail_fputs(bounce,
+			"####################################################################");
+	post_mail_fputs(bounce, "");
+	post_mail_fprintf(bounce,
+		      "Your message could not be delivered for %.1f hours.",
+			  var_delay_warn_time / 3600.0);
+	post_mail_fprintf(bounce,
+			  "It will be retried until it is %.1f days old.",
+			  var_max_queue_time / 86400.0);
+    } else if (bounce_info->report_type == BOUNCE_REPORT_SUCCESS) {
+	post_mail_fputs(bounce,
+			"Your message was sucessfully delivered to the destination(s) listed");
+	post_mail_fputs(bounce,
+			"below. In the case of delivery to mailbox you will receive no further");
+	post_mail_fputs(bounce,
+	    "notifications. In the case of other deliveries you may still");
+	post_mail_fputs(bounce,
+			"receive notifications of mail delivery errors.");
+
+    } else {
+	post_mail_fputs(bounce,
+		"Enclosed is the mail delivery report that you requested.");
+    }
+    if (UNDELIVERED(bounce_info->report_type)) {
+	post_mail_fputs(bounce, "");
+	post_mail_fprintf(bounce,
+			  "For further assistance, please send mail to <%s>",
+			  MAIL_ADDR_POSTMASTER);
+	post_mail_fputs(bounce, "");
+	post_mail_fprintf(bounce,
+	       "If you do so, please include this problem report. You can");
+	post_mail_fprintf(bounce,
+		"delete your own text from the attached returned message.");
+    }
+    post_mail_fputs(bounce, "");
+    post_mail_fprintf(bounce, "\t\t\tThe %s program", var_mail_name);
     return (vstream_ferror(bounce));
 }
 
@@ -530,8 +563,6 @@ static void bounce_print_wrap(VSTREAM *bounce, BOUNCE_INFO *bounce_info,
 
 int     bounce_recipient_log(VSTREAM *bounce, BOUNCE_INFO *bounce_info)
 {
-    RECIPIENT *rcpt = &bounce_info->rcpt_buf->rcpt;
-    DSN    *dsn = &bounce_info->dsn_buf->dsn;
 
     /*
      * Mask control and non-ASCII characters (done in bounce_log_read()),
@@ -539,15 +570,16 @@ int     bounce_recipient_log(VSTREAM *bounce, BOUNCE_INFO *bounce_info)
      * piped into other programs. Sort of like TCP Wrapper's safe_finger
      * program.
      */
-#define NON_NULL_EMPTY(s) ((s) && *(s))
-
     post_mail_fputs(bounce, "");
-    if (NON_NULL_EMPTY(rcpt->orig_addr)) {
+    if (bounce_info->log_handle->rcpt.orig_addr) {
 	bounce_print_wrap(bounce, bounce_info, "<%s> (expanded from <%s>): %s",
-			  rcpt->address, rcpt->orig_addr, dsn->reason);
+			  bounce_info->log_handle->rcpt.address,
+			  bounce_info->log_handle->rcpt.orig_addr,
+			  bounce_info->log_handle->dsn.reason);
     } else {
 	bounce_print_wrap(bounce, bounce_info, "<%s>: %s",
-			  rcpt->address, dsn->reason);
+			  bounce_info->log_handle->rcpt.address,
+			  bounce_info->log_handle->dsn.reason);
     }
     return (vstream_ferror(bounce));
 }
@@ -557,7 +589,6 @@ int     bounce_recipient_log(VSTREAM *bounce, BOUNCE_INFO *bounce_info)
 int     bounce_diagnostic_log(VSTREAM *bounce, BOUNCE_INFO *bounce_info,
 			              int notify_filter)
 {
-    RECIPIENT *rcpt = &bounce_info->rcpt_buf->rcpt;
     int     count = 0;
 
     /*
@@ -571,16 +602,14 @@ int     bounce_diagnostic_log(VSTREAM *bounce, BOUNCE_INFO *bounce_info,
      */
     if (bounce_info->log_handle == 0
 	|| bounce_log_rewind(bounce_info->log_handle)) {
-	if (IS_FAILURE_TEMPLATE(bounce_info->template)) {
-	    post_mail_fputs(bounce, "");
+	if (bounce_info->report_type == BOUNCE_REPORT_FAIL) {
 	    post_mail_fputs(bounce, "\t--- Delivery report unavailable ---");
 	    count = 1;				/* XXX don't abort */
 	}
     } else {
-	while (bounce_log_read(bounce_info->log_handle, bounce_info->rcpt_buf,
-			       bounce_info->dsn_buf) != 0) {
-	    if (rcpt->dsn_notify == 0		/* compat */
-		|| (rcpt->dsn_notify & notify_filter)) {
+	while (bounce_log_read(bounce_info->log_handle) != 0) {
+	    if (bounce_info->log_handle->rcpt.dsn_notify == 0	/* compat */
+	    || (bounce_info->log_handle->rcpt.dsn_notify & notify_filter)) {
 		count++;
 		if (bounce_recipient_log(bounce, bounce_info) != 0)
 		    break;
@@ -615,7 +644,7 @@ int     bounce_header_dsn(VSTREAM *bounce, BOUNCE_INFO *bounce_info)
 #if 0
     post_mail_fprintf(bounce, "Received-From-MTA: dns; %s", "whatever");
 #endif
-    if (NON_NULL_EMPTY(bounce_info->dsn_envid)) {
+    if (bounce_info->dsn_envid) {
 	post_mail_fprintf(bounce, "Original-Envelope-Id: %s",
 			  bounce_info->dsn_envid);
     }
@@ -634,11 +663,9 @@ int     bounce_header_dsn(VSTREAM *bounce, BOUNCE_INFO *bounce_info)
 
 int     bounce_recipient_dsn(VSTREAM *bounce, BOUNCE_INFO *bounce_info)
 {
-    RECIPIENT *rcpt = &bounce_info->rcpt_buf->rcpt;
-    DSN    *dsn = &bounce_info->dsn_buf->dsn;
-
     post_mail_fputs(bounce, "");
-    post_mail_fprintf(bounce, "Final-Recipient: rfc822; %s", rcpt->address);
+    post_mail_fprintf(bounce, "Final-Recipient: rfc822; %s",
+		      bounce_info->log_handle->rcpt.address);
 
     /*
      * XXX DSN
@@ -657,33 +684,39 @@ int     bounce_recipient_dsn(VSTREAM *bounce, BOUNCE_INFO *bounce_info)
      * MTA can't use that same information in its own DSNs.
      * 
      * Postfix always reports an Original-Recipient field, because it is more
-     * more useful and more consistent.
+     * more useful and more inconsistent.
      */
-    if (NON_NULL_EMPTY(rcpt->dsn_orcpt)) {
-	post_mail_fprintf(bounce, "Original-Recipient: %s", rcpt->dsn_orcpt);
-    } else if (NON_NULL_EMPTY(rcpt->orig_addr)) {
+    if (bounce_info->log_handle->rcpt.dsn_orcpt) {
+	post_mail_fprintf(bounce, "Original-Recipient: %s",
+			  bounce_info->log_handle->rcpt.dsn_orcpt);
+    } else if (bounce_info->log_handle->rcpt.orig_addr) {
 	post_mail_fprintf(bounce, "Original-Recipient: rfc822; %s",
-			  rcpt->orig_addr);
+			  bounce_info->log_handle->rcpt.orig_addr);
     }
     post_mail_fprintf(bounce, "Action: %s",
-		      IS_FAILURE_TEMPLATE(bounce_info->template) ?
-		      "failed" : dsn->action);
-    post_mail_fprintf(bounce, "Status: %s", dsn->status);
-    if (NON_NULL_EMPTY(dsn->mtype) && NON_NULL_EMPTY(dsn->mname))
+		      bounce_info->report_type == BOUNCE_REPORT_FAIL ?
+		      "failed" : bounce_info->log_handle->dsn.action);
+    post_mail_fprintf(bounce, "Status: %s",
+		      bounce_info->log_handle->dsn.status);
+    if (bounce_info->log_handle->dsn.mtype
+	&& bounce_info->log_handle->dsn.mname)
 	bounce_print_wrap(bounce, bounce_info, "Remote-MTA: %s; %s",
-			  dsn->mtype, dsn->mname);
-    if (NON_NULL_EMPTY(dsn->dtype) && NON_NULL_EMPTY(dsn->dtext))
+			  bounce_info->log_handle->dsn.mtype,
+			  bounce_info->log_handle->dsn.mname);
+    if (bounce_info->log_handle->dsn.dtype
+	&& bounce_info->log_handle->dsn.dtext)
 	bounce_print_wrap(bounce, bounce_info, "Diagnostic-Code: %s; %s",
-			  dsn->dtype, dsn->dtext);
+			  bounce_info->log_handle->dsn.dtype,
+			  bounce_info->log_handle->dsn.dtext);
     else
 	bounce_print_wrap(bounce, bounce_info, "Diagnostic-Code: X-%s; %s",
-			  bounce_info->mail_name, dsn->reason);
+			  bounce_info->mail_name,
+			  bounce_info->log_handle->dsn.reason);
 #if 0
-    if (dsn->time > 0)
-	post_mail_fprintf(bounce, "Last-Attempt-Date: %s",
-			  mail_date(dsn->time));
+    post_mail_fprintf(bounce, "Last-Attempt-Date: %s",
+		      bounce_info->log_handle->log_time);
 #endif
-    if (IS_DELAY_TEMPLATE(bounce_info->template))
+    if (bounce_info->report_type == BOUNCE_REPORT_WARN)
 	post_mail_fprintf(bounce, "Will-Retry-Until: %s",
 		 mail_date(bounce_info->arrival_time + var_max_queue_time));
     return (vstream_ferror(bounce));
@@ -694,7 +727,6 @@ int     bounce_recipient_dsn(VSTREAM *bounce, BOUNCE_INFO *bounce_info)
 int     bounce_diagnostic_dsn(VSTREAM *bounce, BOUNCE_INFO *bounce_info,
 			              int notify_filter)
 {
-    RECIPIENT *rcpt = &bounce_info->rcpt_buf->rcpt;
     int     count = 0;
 
     /*
@@ -708,13 +740,12 @@ int     bounce_diagnostic_dsn(VSTREAM *bounce, BOUNCE_INFO *bounce_info,
      */
     if (bounce_info->log_handle == 0
 	|| bounce_log_rewind(bounce_info->log_handle)) {
-	if (IS_FAILURE_TEMPLATE(bounce_info->template))
+	if (bounce_info->report_type == BOUNCE_REPORT_FAIL)
 	    count = 1;				/* XXX don't abort */
     } else {
-	while (bounce_log_read(bounce_info->log_handle, bounce_info->rcpt_buf,
-			       bounce_info->dsn_buf) != 0) {
-	    if (rcpt->dsn_notify == 0		/* compat */
-		|| (rcpt->dsn_notify & notify_filter)) {
+	while (bounce_log_read(bounce_info->log_handle) != 0) {
+	    if (bounce_info->log_handle->rcpt.dsn_notify == 0	/* compat */
+	    || (bounce_info->log_handle->rcpt.dsn_notify & notify_filter)) {
 		count++;
 		if (bounce_recipient_dsn(bounce, bounce_info) != 0)
 		    break;
@@ -745,13 +776,10 @@ int     bounce_original(VSTREAM *bounce, BOUNCE_INFO *bounce_info,
     /*
      * MIME headers.
      */
-#define IS_UNDELIVERED_TEMPLATE(template) \
-        (IS_FAILURE_TEMPLATE(template) || IS_DELAY_TEMPLATE(template))
-
     post_mail_fputs(bounce, "");
     post_mail_fprintf(bounce, "--%s", bounce_info->mime_boundary);
     post_mail_fprintf(bounce, "Content-Description: %s%s",
-		      IS_UNDELIVERED_TEMPLATE(bounce_info->template) ?
+		      UNDELIVERED(bounce_info->report_type) ?
 		      "Undelivered " : "",
 		      headers_only == DSN_RET_HDRS ?
 		      "Message Headers" : "Message");
@@ -804,23 +832,22 @@ int     bounce_original(VSTREAM *bounce, BOUNCE_INFO *bounce_info,
 
 void    bounce_delrcpt(BOUNCE_INFO *bounce_info)
 {
-    RECIPIENT *rcpt = &bounce_info->rcpt_buf->rcpt;
-
     if (bounce_info->orig_fp != 0
 	&& bounce_info->log_handle != 0
 	&& bounce_log_rewind(bounce_info->log_handle) == 0)
-	while (bounce_log_read(bounce_info->log_handle, bounce_info->rcpt_buf,
-			       bounce_info->dsn_buf) != 0)
-	    if (rcpt->offset > 0)
-		deliver_completed(bounce_info->orig_fp, rcpt->offset);
+	while (bounce_log_read(bounce_info->log_handle) != 0)
+	    if (bounce_info->log_handle->rcpt.offset > 0)
+		deliver_completed(bounce_info->orig_fp,
+				  bounce_info->log_handle->rcpt.offset);
 }
 
 /* bounce_delrcpt_one - delete one recipient from original queue file */
 
 void    bounce_delrcpt_one(BOUNCE_INFO *bounce_info)
 {
-    RECIPIENT *rcpt = &bounce_info->rcpt_buf->rcpt;
-
-    if (bounce_info->orig_fp != 0 && rcpt->offset > 0)
-	deliver_completed(bounce_info->orig_fp, rcpt->offset);
+    if (bounce_info->orig_fp != 0
+	&& bounce_info->log_handle != 0
+	&& bounce_info->log_handle->rcpt.offset > 0)
+	deliver_completed(bounce_info->orig_fp,
+			  bounce_info->log_handle->rcpt.offset);
 }
