@@ -10,7 +10,7 @@
 /*	CLEANUP_STATE *state;
 /*	int	type;
 /*	const char *buf;
-/*	ssize_t	len;
+/*	int	len;
 /* DESCRIPTION
 /*	This module processes message records with information extracted
 /*	from message content, or with recipients that are stored after the
@@ -64,7 +64,7 @@
 #include <mail_params.h>
 #include <mail_proto.h>
 #include <dsn_mask.h>
-#include <rec_attr_map.h>
+#include <dsn_attr_map.h>
 
 /* Application-specific. */
 
@@ -72,13 +72,13 @@
 
 #define STR(x)	vstring_str(x)
 
-static void cleanup_extracted_process(CLEANUP_STATE *, int, const char *, ssize_t);
+static void cleanup_extracted_process(CLEANUP_STATE *, int, const char *, int);
 static void cleanup_extracted_finish(CLEANUP_STATE *);
 
 /* cleanup_extracted - initialize extracted segment */
 
 void    cleanup_extracted(CLEANUP_STATE *state, int type,
-			          const char *buf, ssize_t len)
+			          const char *buf, int len)
 {
 
     /*
@@ -96,45 +96,20 @@ void    cleanup_extracted(CLEANUP_STATE *state, int type,
 /* cleanup_extracted_process - process one extracted envelope record */
 
 void    cleanup_extracted_process(CLEANUP_STATE *state, int type,
-				          const char *buf, ssize_t len)
+				          const char *buf, int len)
 {
-    const char *myname = "cleanup_extracted_process";
     const char *encoding;
+    const char generated_by_cleanup[] = {
+	REC_TYPE_FILT, REC_TYPE_RDR, REC_TYPE_ATTR,
+	REC_TYPE_RRTO, REC_TYPE_ERTO, 0,
+    };
     char   *attr_name;
     char   *attr_value;
     const char *error_text;
-    int     extra_opts;
     int     junk;
 
-#ifdef DELAY_ACTION
-    int     defer_delay;
-
-#endif
-
     if (msg_verbose)
-	msg_info("extracted envelope %c %.*s", type, (int) len, buf);
-
-    if (type == REC_TYPE_FLGS) {
-	/* Not part of queue file format. */
-	extra_opts = atoi(buf);
-	if (extra_opts & ~CLEANUP_FLAG_MASK_EXTRA)
-	    msg_warn("%s: ignoring bad extra flags: 0x%x",
-		     state->queue_id, extra_opts);
-	else
-	    state->flags |= extra_opts;
-	return;
-    }
-#ifdef DELAY_ACTION
-    if (type == REC_TYPE_DELAY) {
-	/* Not part of queue file format. */
-	defer_delay = atoi(buf);
-	if (defer_delay <= 0)
-	    msg_warn("%s: ignoring bad delay time: %s", state->queue_id, buf);
-	else
-	    state->defer_delay = defer_delay;
-	return;
-    }
-#endif
+	msg_info("extracted envelope %c %.*s", type, len, buf);
 
     if (strchr(REC_TYPE_EXTRACT, type) == 0) {
 	msg_warn("%s: message rejected: "
@@ -151,24 +126,24 @@ void    cleanup_extracted_process(CLEANUP_STATE *state, int type,
      * an upgrade without losing mail.
      */
     if (type == REC_TYPE_ATTR) {
-	vstring_strcpy(state->attr_buf, buf);
-	error_text = split_nameval(STR(state->attr_buf), &attr_name, &attr_value);
-	if (error_text != 0) {
-	    msg_warn("%s: message rejected: malformed attribute: %s: %.100s",
-		     state->queue_id, error_text, buf);
-	    state->errs |= CLEANUP_STAT_BAD;
-	    return;
-	}
-	/* Zero-length values are place holders for unavailable values. */
-	if (*attr_value == 0) {
-	    msg_warn("%s: spurious null attribute value for \"%s\" -- ignored",
-		     state->queue_id, attr_name);
-	    return;
-	}
-	if ((junk = rec_attr_map(attr_name)) != 0) {
-	    buf = attr_value;
-	    type = junk;
-	}
+        vstring_strcpy(state->attr_buf, buf);
+        error_text = split_nameval(STR(state->attr_buf), &attr_name, &attr_value);
+        if (error_text != 0) {
+            msg_warn("%s: message rejected: malformed attribute: %s: %.100s",
+                     state->queue_id, error_text, buf);
+            state->errs |= CLEANUP_STAT_BAD;
+            return;
+        }
+        /* Zero-length values are place holders for unavailable values. */
+        if (*attr_value == 0) {
+            msg_warn("%s: spurious null attribute value for \"%s\" -- ignored",
+                     state->queue_id, attr_name);
+            return;
+        }
+        if ((junk = dsn_attr_map(attr_name)) != 0) {
+            buf = attr_value;
+            type = junk;
+        }
     }
 
     /*
@@ -200,10 +175,6 @@ void    cleanup_extracted_process(CLEANUP_STATE *state, int type,
 	if (state->orig_rcpt == 0)
 	    state->orig_rcpt = mystrdup(buf);
 	cleanup_addr_recipient(state, buf);
-	if (cleanup_milters != 0
-	    && state->milters == 0
-	    && CLEANUP_MILTER_OK(state))
-	    cleanup_milter_emul_rcpt(state, cleanup_milters, buf);
 	myfree(state->orig_rcpt);
 	state->orig_rcpt = 0;
 	if (state->dsn_orcpt != 0) {
@@ -213,7 +184,7 @@ void    cleanup_extracted_process(CLEANUP_STATE *state, int type,
 	state->dsn_notify = 0;
 	return;
     }
-    if (type == REC_TYPE_DONE || type == REC_TYPE_DRCP) {
+    if (type == REC_TYPE_DONE) {
 	if (state->orig_rcpt != 0) {
 	    myfree(state->orig_rcpt);
 	    state->orig_rcpt = 0;
@@ -258,15 +229,6 @@ void    cleanup_extracted_process(CLEANUP_STATE *state, int type,
 	return;
     }
     if (type == REC_TYPE_END) {
-	/* Make room to append recipient. */
-	if ((state->milters || cleanup_milters)
-	    && state->append_rcpt_pt_offset < 0) {
-	    if ((state->append_rcpt_pt_offset = vstream_ftell(state->dst)) < 0)
-		msg_fatal("%s: vstream_ftell %s: %m:", myname, cleanup_path);
-	    cleanup_out_format(state, REC_TYPE_PTR, REC_TYPE_PTR_FORMAT, 0L);
-	    if ((state->append_rcpt_pt_target = vstream_ftell(state->dst)) < 0)
-		msg_fatal("%s: vstream_ftell %s: %m:", myname, cleanup_path);
-	}
 	state->flags &= ~CLEANUP_FLAG_INRCPT;
 	state->flags |= CLEANUP_FLAG_END_SEEN;
 	cleanup_extracted_finish(state);
@@ -279,11 +241,17 @@ void    cleanup_extracted_process(CLEANUP_STATE *state, int type,
     if (state->flags & CLEANUP_FLAG_INRCPT)
 	/* Tell qmgr that recipient records are mixed with other information. */
 	state->qmgr_opts |= QMGR_READ_FLAG_MIXED_RCPT_OTHER;
-    cleanup_out(state, type, buf, len);
-    return;
+    if (strchr(generated_by_cleanup, type) != 0) {
+	/* Use our own header/body info instead. */
+	return;
+    } else {
+	/* Pass on other non-recipient record. */
+	cleanup_out(state, type, buf, len);
+	return;
+    }
 }
 
-/* cleanup_extracted_finish - complete the third message segment */
+/* cleanup_extracted_finish - process one extracted envelope record */
 
 void    cleanup_extracted_finish(CLEANUP_STATE *state)
 {
@@ -300,4 +268,32 @@ void    cleanup_extracted_finish(CLEANUP_STATE *state)
      * Terminate the extracted segment.
      */
     cleanup_out_string(state, REC_TYPE_END, "");
+
+    /*
+     * vstream_fseek() would flush the buffer anyway, but the code just reads
+     * better if we flush first, because it makes seek error handling more
+     * straightforward.
+     */
+    if (vstream_fflush(state->dst)) {
+	if (errno == EFBIG) {
+	    msg_warn("%s: queue file size limit exceeded", state->queue_id);
+	    state->errs |= CLEANUP_STAT_SIZE;
+	} else {
+	    msg_warn("%s: write queue file: %m", state->queue_id);
+	    state->errs |= CLEANUP_STAT_WRITE;
+	}
+	return;
+    }
+
+    /*
+     * Update the preliminary message size and count fields with the actual
+     * values.
+     */
+    if (vstream_fseek(state->dst, 0L, SEEK_SET) < 0)
+	msg_fatal("%s: vstream_fseek %s: %m", myname, cleanup_path);
+    cleanup_out_format(state, REC_TYPE_SIZE, REC_TYPE_SIZE_FORMAT,
+	    (REC_TYPE_SIZE_CAST1) (state->xtra_offset - state->data_offset),
+		       (REC_TYPE_SIZE_CAST2) state->data_offset,
+		       (REC_TYPE_SIZE_CAST3) state->rcpt_count,
+		       (REC_TYPE_SIZE_CAST4) state->qmgr_opts);
 }
