@@ -134,6 +134,7 @@ static int deliver_message(DELIVER_REQUEST *request)
     RECIPIENT *rcpt;
     int     nrcpt;
     DSN_SPLIT dp;
+    DSN     dsn;
 
     if (msg_verbose)
 	msg_info("deliver_message: from %s", request->sender);
@@ -166,13 +167,12 @@ static int deliver_message(DELIVER_REQUEST *request)
 #define BOUNCE_FLAGS(request) DEL_REQ_TRACE_FLAGS(request->flags)
 
     dsn_split(&dp, "2.0.0", request->nexthop);
+    (void) DSN_SIMPLE(&dsn, DSN_STATUS(dp.dsn), dp.text);
     for (nrcpt = 0; nrcpt < request->rcpt_list.len; nrcpt++) {
 	rcpt = request->rcpt_list.info + nrcpt;
 	if (rcpt->offset >= 0) {
 	    status = sent(BOUNCE_FLAGS(request), request->queue_id,
-			  rcpt->orig_addr, rcpt->address, rcpt->offset,
-			  "none", DSN_CODE(dp.dsn),
-			  request->arrival_time, "%s", dp.text);
+			  request->arrival_time, rcpt, "none", &dsn);
 	    if (status == 0 && (request->flags & DEL_REQ_FLAG_SUCCESS))
 		deliver_completed(src, rcpt->offset);
 	    result |= status;
