@@ -37,8 +37,7 @@
 /*	number of CNAME indirections. All result names (including
 /*	null terminator) will fit a buffer of size DNS_NAME_LEN.
 /*	All name results are validated by \fIvalid_hostname\fR();
-/*	an invalid name is reported as a DNS_INVAL result, while
-/*	malformed replies are reported as transient errors.
+/*	an invalid name is reported as a transient error.
 /*
 /*	dns_lookup_l() and dns_lookup_v() allow the user to specify
 /*	a list of resource types.
@@ -47,8 +46,6 @@
 /* .fi
 /* .IP name
 /*	The name to be looked up in the domain name system.
-/*	This name must pass the valid_hostname() test; it
-/*	must not be an IP address.
 /* .IP type
 /*	The resource record type to be looked up (T_A, T_MX etc.).
 /* .IP rflags
@@ -62,18 +59,16 @@
 /*	Append local domain to unqualified names.
 /* .RE
 /* .IP lflags
-/*	Multi-type request control for dns_lookup_l() and dns_lookup_v().
-/*	For convenience, DNS_REQ_FLAG_NONE requests no special
-/*	processing. Invoke dns_lookup() for all specified resource
-/*	record types in the specified order, and merge their results.
-/*	Otherwise, specify one or more of the following:
+/*	Multi-type request control for dns_lookup_l() and
+/*	dns_lookup_v(). This is one of the following:
 /* .RS
-/* .IP DNS_REQ_FLAG_STOP_INVAL
-/*	Invoke dns_lookup() for the resource types in the order as
-/*	specified, and return when dns_lookup() returns DNS_INVAL.
-/* .IP DNS_REQ_FLAG_STOP_OK
-/*	Invoke dns_lookup() for the resource types in the order as
-/*	specified, and return when dns_lookup() returns DNS_OK.
+/* .IP DNS_REQ_FLAG_ANY
+/*	Call dns_lookup() for each specified resource record type
+/*	in the specified order, until the list is exhausted or
+/*	until some result is DNS_OK.
+/* .IP DNS_REQ_FLAG_ALL
+/*	Call dns_lookup() for all specified resource record types
+/*	in the specified order, and merge their results.
 /* .RE
 /* .IP ltype
 /*	The resource record types to be looked up. In the case of
@@ -98,11 +93,8 @@
 /*	The DNS query succeeded.
 /* .IP DNS_NOTFOUND
 /*	The DNS query succeeded; the requested information was not found.
-/* .IP DNS_INVAL
-/*	The DNS query succeeded; the result failed the valid_hostname() test.
 /* .IP DNS_RETRY
-/*	The query failed, or the reply was malformed.
-/*	The problem is considered transient.
+/*	The query failed; the problem is transient.
 /* .IP DNS_FAIL
 /*	The query failed.
 /* BUGS
@@ -150,12 +142,10 @@
  /*
   * Structure to keep track of things while decoding a name server reply.
   */
-#define DEF_DNS_REPLY_SIZE	4096	/* in case we're using TCP */
-#define MAX_DNS_REPLY_SIZE	32768	/* in case we're using TCP */
+#define DNS_REPLY_SIZE	4096		/* in case we're using TCP */
 
 typedef struct DNS_REPLY {
-    unsigned char *buf;			/* raw reply data */
-    size_t  buf_len;			/* reply buffer length */
+    unsigned char buf[DNS_REPLY_SIZE];	/* raw reply data */
     int     query_count;		/* number of queries */
     int     answer_count;		/* number of answers */
     unsigned char *query_start;		/* start of query data */
@@ -175,14 +165,6 @@ static int dns_query(const char *name, int type, int flags,
     int     len;
     unsigned long saved_options;
 
-    /*
-     * Initialize the reply buffer.
-     */
-    if (reply->buf == 0) {
-	reply->buf = (unsigned char *) mymalloc(DEF_DNS_REPLY_SIZE);
-	reply->buf_len = DEF_DNS_REPLY_SIZE;
-    }
-
     /*
      * Initialize the name service.
      */
@@ -201,60 +183,53 @@ static int dns_query(const char *name, int type, int flags,
     if ((flags & USER_FLAGS) != flags)
 	msg_panic("dns_query: bad flags: %d", flags);
     saved_options = (_res.options & USER_FLAGS);
+    _res.options &= ~saved_options;
+    _res.options |= flags;
 
     /*
      * Perform the lookup. Claim that the information cannot be found if and
      * only if the name server told us so.
      */
-    for (;;) {
-	_res.options &= ~saved_options;
-	_res.options |= flags;
-	len = res_search((char *) name, C_IN, type, reply->buf, reply->buf_len);
-	_res.options &= ~flags;
-	_res.options |= saved_options;
-	if (len < 0) {
-	    if (why)
-		vstring_sprintf(why, "Host or domain name not found. "
-				"Name service error for name=%s type=%s: %s",
+    len = res_search((char *) name, C_IN, type, reply->buf, sizeof(reply->buf));
+    _res.options &= ~flags;
+    _res.options |= saved_options;
+    if (len < 0) {
+	if (why)
+	    vstring_sprintf(why, "Host or domain name not found. "
+			    "Name service error for name=%s type=%s: %s",
 			    name, dns_strtype(type), dns_strerror(h_errno));
-	    if (msg_verbose)
-		msg_info("dns_query: %s (%s): %s",
-			 name, dns_strtype(type), dns_strerror(h_errno));
-	    switch (h_errno) {
-	    case NO_RECOVERY:
-		return (DNS_FAIL);
-	    case HOST_NOT_FOUND:
-	    case NO_DATA:
-		return (DNS_NOTFOUND);
-	    default:
-		return (DNS_RETRY);
-	    }
-	}
 	if (msg_verbose)
-	    msg_info("dns_query: %s (%s): OK", name, dns_strtype(type));
-
-	reply_header = (HEADER *) reply->buf;
-	if (reply_header->tc == 0 || reply->buf_len >= MAX_DNS_REPLY_SIZE)
-	    break;
-	reply->buf = (unsigned char *)
-	    myrealloc((char *) reply->buf, 2 * reply->buf_len);
-	reply->buf_len *= 2;
+	    msg_info("dns_query: %s (%s): %s",
+		     name, dns_strtype(type), dns_strerror(h_errno));
+	switch (h_errno) {
+	case NO_RECOVERY:
+	    return (DNS_FAIL);
+	case HOST_NOT_FOUND:
+	case NO_DATA:
+	    return (DNS_NOTFOUND);
+	default:
+	    return (DNS_RETRY);
+	}
     }
+    if (msg_verbose)
+	msg_info("dns_query: %s (%s): OK", name, dns_strtype(type));
 
     /*
      * Paranoia.
      */
-    if (len > reply->buf_len) {
+    if (len > sizeof(reply->buf)) {
 	msg_warn("reply length %d > buffer length %d for name=%s type=%s",
-		 len, (int) reply->buf_len, name, dns_strtype(type));
-	len = reply->buf_len;
+		 len, (int) sizeof(reply->buf), name, dns_strtype(type));
+	len = sizeof(reply->buf);
     }
 
     /*
      * Initialize the reply structure. Some structure members are filled on
      * the fly while the reply is being parsed.
      */
-    reply->end = reply->buf + len;
+    if ((reply->end = reply->buf + len) > reply->buf + sizeof(reply->buf))
+	reply->end = reply->buf + sizeof(reply->buf);
+    reply_header = (HEADER *) reply->buf;
     reply->query_start = reply->buf + sizeof(HEADER);
     reply->answer_start = 0;
     reply->query_count = ntohs(reply_header->qdcount);
@@ -349,12 +324,11 @@ static int valid_rr_name(const char *name, const char *location,
 
 /* dns_get_rr - extract resource record from name server reply */
 
-static int dns_get_rr(DNS_RR **list, const char *orig_name, DNS_REPLY *reply,
-		              unsigned char *pos, char *rr_name,
-		              DNS_FIXED *fixed)
+static DNS_RR *dns_get_rr(DNS_REPLY *reply, unsigned char *pos,
+			          char *rr_name, DNS_FIXED *fixed)
 {
     char    temp[DNS_NAME_LEN];
-    ssize_t data_len;
+    int     data_len;
     unsigned pref = 0;
     unsigned char *src;
     unsigned char *dst;
@@ -362,9 +336,8 @@ static int dns_get_rr(DNS_RR **list, const char *orig_name, DNS_REPLY *reply,
 
 #define MIN2(a, b)	((unsigned)(a) < (unsigned)(b) ? (a) : (b))
 
-    *list = 0;
     if (pos + fixed->length > reply->end)
-	return (DNS_RETRY);
+	return (0);
 
     switch (fixed->type) {
     default:
@@ -377,23 +350,23 @@ static int dns_get_rr(DNS_RR **list, const char *orig_name, DNS_REPLY *reply,
     case T_NS:
     case T_PTR:
 	if (dn_expand(reply->buf, reply->end, pos, temp, sizeof(temp)) < 0)
-	    return (DNS_RETRY);
+	    return (0);
 	if (!valid_rr_name(temp, "resource data", fixed->type, reply))
-	    return (DNS_INVAL);
+	    return (0);
 	data_len = strlen(temp) + 1;
 	break;
     case T_MX:
 	GETSHORT(pref, pos);
 	if (dn_expand(reply->buf, reply->end, pos, temp, sizeof(temp)) < 0)
-	    return (DNS_RETRY);
+	    return (0);
 	if (!valid_rr_name(temp, "resource data", fixed->type, reply))
-	    return (DNS_INVAL);
+	    return (0);
 	data_len = strlen(temp) + 1;
 	break;
     case T_A:
 	if (fixed->length != INET_ADDR_LEN) {
 	    msg_warn("extract_answer: bad address length: %d", fixed->length);
-	    return (DNS_RETRY);
+	    return (0);
 	}
 	if (fixed->length > sizeof(temp))
 	    msg_panic("dns_get_rr: length %d > DNS_NAME_LEN",
@@ -405,7 +378,7 @@ static int dns_get_rr(DNS_RR **list, const char *orig_name, DNS_REPLY *reply,
     case T_AAAA:
 	if (fixed->length != INET6_ADDR_LEN) {
 	    msg_warn("extract_answer: bad address length: %d", fixed->length);
-	    return (DNS_RETRY);
+	    return (0);
 	}
 	if (fixed->length > sizeof(temp))
 	    msg_panic("dns_get_rr: length %d > DNS_NAME_LEN",
@@ -424,9 +397,8 @@ static int dns_get_rr(DNS_RR **list, const char *orig_name, DNS_REPLY *reply,
 	*dst = 0;
 	break;
     }
-    *list = dns_rr_create(orig_name, rr_name, fixed->type, fixed->class,
-			  fixed->ttl, pref, temp, data_len);
-    return (DNS_OK);
+    return (dns_rr_create(rr_name, fixed->type, fixed->class, fixed->ttl,
+			  pref, temp, data_len));
 }
 
 /* dns_get_alias - extract CNAME from name server reply */
@@ -439,13 +411,13 @@ static int dns_get_alias(DNS_REPLY *reply, unsigned char *pos,
     if (dn_expand(reply->buf, reply->end, pos, cname, c_len) < 0)
 	return (DNS_RETRY);
     if (!valid_rr_name(cname, "resource data", fixed->type, reply))
-	return (DNS_INVAL);
+	return (DNS_RETRY);
     return (DNS_OK);
 }
 
 /* dns_get_answer - extract answers from name server reply */
 
-static int dns_get_answer(const char *orig_name, DNS_REPLY *reply, int type,
+static int dns_get_answer(DNS_REPLY *reply, int type,
 	             DNS_RR **rrlist, VSTRING *fqdn, char *cname, int c_len)
 {
     char    rr_name[DNS_NAME_LEN];
@@ -456,15 +428,14 @@ static int dns_get_answer(const char *orig_name, DNS_REPLY *reply, int type,
     DNS_RR *rr;
     int     resource_found = 0;
     int     cname_found = 0;
-    int     not_found_status = DNS_NOTFOUND;	/* can't happen */
-    int     status;
+    int     not_found_status = DNS_NOTFOUND;
 
     /*
      * Initialize. Skip over the name server query if we haven't yet.
      */
     if (reply->answer_start == 0)
-	if ((status = dns_skip_query(reply)) < 0)
-	    return (status);
+	if (dns_skip_query(reply) < 0)
+	    return (DNS_RETRY);
     pos = reply->answer_start;
     if (rrlist)
 	*rrlist = 0;
@@ -473,12 +444,12 @@ static int dns_get_answer(const char *orig_name, DNS_REPLY *reply, int type,
      * Either this, or use a GOTO for emergency exits. The purpose is to
      * prevent incomplete answers from being passed back to the caller.
      */
-#define CORRUPT(status) { \
+#define CORRUPT { \
 	if (rrlist && *rrlist) { \
 	    dns_rr_free(*rrlist); \
 	    *rrlist = 0; \
 	} \
-	return (status); \
+	return (DNS_RETRY); \
     }
 
     /*
@@ -490,21 +461,21 @@ static int dns_get_answer(const char *orig_name, DNS_REPLY *reply, int type,
 	 * Optionally extract the fully-qualified domain name.
 	 */
 	if (pos >= reply->end)
-	    CORRUPT(DNS_RETRY);
+	    CORRUPT;
 	len = dn_expand(reply->buf, reply->end, pos, rr_name, DNS_NAME_LEN);
 	if (len < 0)
-	    CORRUPT(DNS_RETRY);
+	    CORRUPT;
 	pos += len;
 
 	/*
 	 * Extract the fixed reply data: type, class, ttl, length.
 	 */
 	if (pos + RRFIXEDSZ > reply->end)
-	    CORRUPT(DNS_RETRY);
-	if ((status = dns_get_fixed(pos, &fixed)) != DNS_OK)
-	    CORRUPT(status);
+	    CORRUPT;
+	if (dns_get_fixed(pos, &fixed) != DNS_OK)
+	    CORRUPT;
 	if (!valid_rr_name(rr_name, "resource name", fixed.type, reply))
-	    CORRUPT(DNS_INVAL);
+	    CORRUPT;
 	if (fqdn)
 	    vstring_strcpy(fqdn, rr_name);
 	if (msg_verbose)
@@ -516,22 +487,21 @@ static int dns_get_answer(const char *orig_name, DNS_REPLY *reply, int type,
 	 * Optionally extract the requested resource or CNAME data.
 	 */
 	if (pos + fixed.length > reply->end)
-	    CORRUPT(DNS_RETRY);
+	    CORRUPT;
 	if (type == fixed.type || type == T_ANY) {	/* requested type */
 	    if (rrlist) {
-		if ((status = dns_get_rr(&rr, orig_name, reply, pos, rr_name,
-					 &fixed)) == DNS_OK) {
+		if ((rr = dns_get_rr(reply, pos, rr_name, &fixed)) != 0) {
 		    resource_found++;
 		    *rrlist = dns_rr_append(*rrlist, rr);
-		} else if (not_found_status != DNS_RETRY)
-		    not_found_status = status;
+		} else
+		    not_found_status = DNS_RETRY;
 	    } else
 		resource_found++;
 	} else if (fixed.type == T_CNAME) {	/* cname resource */
 	    cname_found++;
 	    if (cname && c_len > 0)
-		if ((status = dns_get_alias(reply, pos, &fixed, cname, c_len)) != DNS_OK)
-		    CORRUPT(status);
+		if (dns_get_alias(reply, pos, &fixed, cname, c_len) != DNS_OK)
+		    CORRUPT;
 	}
 	pos += fixed.length;
     }
@@ -555,10 +525,9 @@ int     dns_lookup(const char *name, unsigned type, unsigned flags,
 {
     char    cname[DNS_NAME_LEN];
     int     c_len = sizeof(cname);
-    static DNS_REPLY reply;
+    DNS_REPLY reply;
     int     count;
     int     status;
-    const char *orig_name = name;
 
     /*
      * DJBDNS produces a bogus A record when given a numerical hostname.
@@ -600,16 +569,15 @@ int     dns_lookup(const char *name, unsigned type, unsigned flags,
 	 * Extract resource records of the requested type. Pick up CNAME
 	 * information just in case the requested data is not found.
 	 */
-	status = dns_get_answer(orig_name, &reply, type, rrlist, fqdn,
-				cname, c_len);
+	status = dns_get_answer(&reply, type, rrlist, fqdn, cname, c_len);
 	switch (status) {
 	default:
 	    if (why)
 		vstring_sprintf(why, "Name service error for name=%s type=%s: "
-				"Malformed or unexpected name server reply",
+				"Malformed name server reply",
 				name, dns_strtype(type));
-	    /* FALLTHROUGH */
 	case DNS_OK:
+	case DNS_NOTFOUND:
 	    return (status);
 	case DNS_RECURSE:
 	    if (msg_verbose)
@@ -648,10 +616,7 @@ int     dns_lookup_l(const char *name, unsigned flags, DNS_RR **rrlist,
 	    non_err = 1;
 	    if (rrlist)
 		*rrlist = dns_rr_append(*rrlist, rr);
-	    if (lflags & DNS_REQ_FLAG_STOP_OK)
-		break;
-	} else if (status == DNS_INVAL) {
-	    if (lflags & DNS_REQ_FLAG_STOP_INVAL)
+	    if (lflags == DNS_REQ_FLAG_ANY)
 		break;
 	} else if (status == DNS_RETRY) {
 	    soft_err = 1;
@@ -685,10 +650,7 @@ int     dns_lookup_v(const char *name, unsigned flags, DNS_RR **rrlist,
 	    non_err = 1;
 	    if (rrlist)
 		*rrlist = dns_rr_append(*rrlist, rr);
-	    if (lflags & DNS_REQ_FLAG_STOP_OK)
-		break;
-	} else if (status == DNS_INVAL) {
-	    if (lflags & DNS_REQ_FLAG_STOP_INVAL)
+	    if (lflags == DNS_REQ_FLAG_ANY)
 		break;
 	} else if (status == DNS_RETRY) {
 	    soft_err = 1;
