@@ -17,6 +17,14 @@
 #include <vstream.h>
 #include <vstring.h>
 
+ /*
+  * Global library.
+  */
+#include <recipient_list.h>
+#include <rcpt_buf.h>
+#include <dsn_buf.h>
+#include <dsn.h>
+
  /*
   * External interface.
   */
@@ -24,23 +32,23 @@ typedef struct {
     /* Private. */
     VSTREAM *fp;			/* open file */
     VSTRING *buf;			/* I/O buffer */
+    RCPT_BUF *rcpt_buf;			/* recipient info */
+    DSN_BUF *dsn_buf;			/* delivery status */
+    char    *compat_status;		/* old logfile compatibility */
+    char    *compat_action;		/* old logfile compatibility */
     /* Public. */
-    const char *recipient;		/* final recipient */
-    const char *status;			/* recipient status */
-    const char *text;			/* why undeliverable */
-    long    offset;			/* start of current record */
+    RECIPIENT rcpt;			/* recipient info */
+    DSN     dsn;			/* delivery status */
 } BOUNCE_LOG;
 
 extern BOUNCE_LOG *bounce_log_open(const char *, const char *, int, int);
 extern BOUNCE_LOG *bounce_log_read(BOUNCE_LOG *);
 extern BOUNCE_LOG *bounce_log_delrcpt(BOUNCE_LOG *);
-extern BOUNCE_LOG *bounce_log_forge(const char *, const char *, const char *);
+extern BOUNCE_LOG *bounce_log_forge(RECIPIENT *, DSN *);
 extern int bounce_log_close(BOUNCE_LOG *);
 
 #define bounce_log_rewind(bp) vstream_fseek((bp)->fp, 0L, SEEK_SET)
 
-#define BOUNCE_LOG_STAT_DELETED	'D'	/* deleted record */
-
 /* LICENSE
 /* .ad
 /* .fi
