@@ -114,6 +114,7 @@ extern char *var_mydest;
 extern char *var_myhostname;
 
 #define VAR_MYDOMAIN		"mydomain"	/* my domain name */
+#define DEF_MYDOMAIN		"localdomain"
 extern char *var_mydomain;
 
  /*
@@ -2350,6 +2351,11 @@ extern char *var_smtp_ehlo_dis_words;
 #define DEF_SMTP_EHLO_DIS_MAPS		""
 extern char *var_smtp_ehlo_dis_maps;
 
+ /*
+  * gcc workaround for warnings about empty or null format strings.
+  */
+extern const char null_format_string[1];
+
 /* LICENSE
 /* .ad
 /* .fi
