@@ -85,18 +85,11 @@
 /*	for testing address rewriting and routing configurations.
 /* .sp
 /*	This feature is available in Postfix version 2.1 and later.
-/* .IP "\fB-C \fIconfig_file\fR"
-/* .IP "\fB-C \fIconfig_dir\fR"
-/*	The path name of the Postfix \fBmain.cf\fR file, or of its
-/*	parent directory. This information is ignored with Postfix
-/*	versions before 2.3.
-/*
-/*	With all Postfix versions, you can specify a directory pathname
-/*	with the MAIL_CONFIG environment variable to override the
-/*	location of configuration files.
+/* .IP "\fB-C \fIconfig_file\fR (ignored)"
+/*	The path name of the \fBsendmail.cf\fR file. Postfix configuration
+/*	files are kept in the \fB/etc/postfix\fR directory.
 /* .IP "\fB-F \fIfull_name\fR
-/*	Set the sender full name. This overrides the NAME environment
-/*	variable, and is used only with messages that
+/*	Set the sender full name. This is used only with messages that
 /*	have no \fBFrom:\fR message header.
 /* .IP "\fB-f \fIsender\fR"
 /*	Set the envelope sender address. This is the address where
@@ -191,7 +184,7 @@
 /*	support DSN.
 /*
 /*	This feature is available in Postfix 2.3 and later.
-/* .IP "\fB-XV\fR (Postfix 2.2 and earlier: \fB-V\fR)"
+/* .IP "\fB-V\fR (with Postfix 2.3 use \fB-XV\fR)"
 /*	Variable Envelope Return Path. Given an envelope sender address
 /*	of the form \fIowner-listname\fR@\fIorigin\fR, each recipient
 /*	\fIuser\fR@\fIdomain\fR receives mail with a personalized envelope
@@ -201,10 +194,28 @@
 /*	\fIowner-listname\fB+\fIuser\fB=\fIdomain\fR@\fIorigin\fR. The default
 /*	\fB+\fR and \fB=\fR characters are configurable with the
 /*	\fBdefault_verp_delimiters\fR configuration parameter.
-/* .IP "\fB-XV\fIxy\fR (Postfix 2.2 and earlier: \fB-V\fIxy\fR)"
-/*	As \fB-XV\fR, but uses \fIx\fR and \fIy\fR as the VERP delimiter
+/* .IP "\fB-V\fIxy\fR (with Postfix 2.3 use \fB-XV\fIxy\fR)"
+/*	As \fB-V\fR, but uses \fIx\fR and \fIy\fR as the VERP delimiter
 /*	characters, instead of the characters specified with the
 /*	\fBdefault_verp_delimiters\fR configuration parameter.
+/* .IP \fB-XV\fR
+/*	Variable Envelope Return Path. Given an envelope sender address
+/*	of the form \fIowner-listname\fR@\fIorigin\fR, each recipient
+/*	\fIuser\fR@\fIdomain\fR receives mail with a personalized envelope
+/*	sender address.
+/* .sp
+/*	By default, the personalized envelope sender address is
+/*	\fIowner-listname\fB+\fIuser\fB=\fIdomain\fR@\fIorigin\fR. The default
+/*	\fB+\fR and \fB=\fR characters are configurable with the
+/*	\fBdefault_verp_delimiters\fR configuration parameter.
+/* .sp
+/*	This feature is available in Postfix version 2.3 and later.
+/* .IP \fB-XV\fIxy\fR
+/*	As \fB-V\fR, but uses \fIx\fR and \fIy\fR as the VERP delimiter
+/*	characters, instead of the characters specified with the
+/*	\fBdefault_verp_delimiters\fR configuration parameter.
+/* .sp
+/*	This feature is available in Postfix version 2.3 and later.
 /* .IP \fB-v\fR
 /*	Send an email report of the first delivery attempt (Postfix
 /*	versions 2.1 and later). Mail delivery
@@ -228,15 +239,11 @@
 /* .fi
 /* .IP \fBMAIL_CONFIG\fR
 /*	Directory with Postfix configuration files.
-/* .IP "\fBMAIL_VERBOSE\fR (value does not matter)"
+/* .IP \fBMAIL_VERBOSE\fR
 /*	Enable verbose logging for debugging purposes.
-/* .IP "\fBMAIL_DEBUG\fR (value does not matter)"
+/* .IP \fBMAIL_DEBUG\fR
 /*	Enable debugging with an external command, as specified with the
 /*	\fBdebugger_command\fR configuration parameter.
-/* .IP \fBNAME\fR
-/*	The sender full name. This is used only with messages that
-/*	have no \fBFrom:\fR message header. See also the \fB-F\fR
-/*	option above.
 /* CONFIGURATION PARAMETERS
 /* .ad
 /* .fi
@@ -326,8 +333,8 @@
 /* .IP "\fBenable_errors_to (no)\fR"
 /*	Report mail delivery errors to the address specified with the
 /*	non-standard Errors-To: message header, instead of the envelope
-/*	sender address (this feature is removed with Postfix version 2.2, is
-/*	turned off by default with Postfix version 2.1, and is always turned on
+/*	sender address (this feature is removed with Postfix 2.2, is
+/*	turned off by default with Postfix 2.1, and is always turned on
 /*	with older Postfix versions).
 /* .IP "\fBmail_owner (postfix)\fR"
 /*	The UNIX system account that owns the Postfix queue and most Postfix
@@ -343,6 +350,9 @@
 /* .IP "\fBsyslog_name (postfix)\fR"
 /*	The mail system name that is prepended to the process name in syslog
 /*	records, so that "smtpd" becomes, for example, "postfix/smtpd".
+/* .IP "\fBtrigger_timeout (10s)\fR"
+/*	The time limit for sending a trigger to a Postfix daemon (for
+/*	example, the \fBpickup\fR(8) or \fBqmgr\fR(8) daemon).
 /* FILES
 /*	/var/spool/postfix, mail queue
 /*	/etc/postfix, configuration files
@@ -488,7 +498,7 @@ static CONFIG_STR_TABLE str_table[] = {
 
 /* output_text - output partial or complete text line */
 
-static void output_text(void *context, int rec_type, const char *buf, ssize_t len,
+static void output_text(void *context, int rec_type, const char *buf, int len,
 			        off_t unused_offset)
 {
     SM_STATE *state = (SM_STATE *) context;
@@ -513,7 +523,7 @@ static void output_header(void *context, int header_class,
     char   *start;
     char   *line;
     char   *next_line;
-    ssize_t len;
+    int     len;
 
     /*
      * Parse the header line, and save copies of recipient addresses in the
@@ -885,13 +895,6 @@ static void enqueue(const int flags, const char *encoding,
     myfree(saved_sender);
 }
 
-/* tempfail - sanitize exit status after library run-time error */
-
-static void tempfail(void)
-{
-    exit(EX_TEMPFAIL);
-}
-
 /* main - the main program */
 
 int     main(int argc, char **argv)
@@ -913,10 +916,9 @@ int     main(int argc, char **argv)
     char   *qtime = 0;
     const char *errstr;
     uid_t   uid;
-    const char *rewrite_context = MAIL_ATTR_RWR_LOCAL;
+    char   *rewrite_context = MAIL_ATTR_RWR_LOCAL;
     int     dsn_notify = 0;
     const char *dsn_envid = 0;
-    int     saved_optind;
 
     /*
      * Be consistent with file permissions.
@@ -959,7 +961,6 @@ int     main(int argc, char **argv)
     if ((slash = strrchr(argv[0], '/')) != 0 && slash[1])
 	argv[0] = slash + 1;
     msg_vstream_init(argv[0], VSTREAM_ERR);
-    msg_cleanup(tempfail);
     msg_syslog_init(mail_task("sendmail"), LOG_PID, LOG_FACILITY);
     set_mail_conf_str(VAR_PROCNAME, var_procname = mystrdup(argv[0]));
 
@@ -976,31 +977,8 @@ int     main(int argc, char **argv)
     }
 
     /*
-     * Further initialization. Load main.cf first, so that command-line
-     * options can override main.cf settings. Pre-scan the argument list so
-     * that we load the right main.cf file.
+     * Further initialization...
      */
-#define GETOPT_LIST "A:B:C:F:GIL:N:O:R:UV:X:b:ce:f:h:imno:p:r:q:tvx"
-
-    saved_optind = optind;
-    while (argv[OPTIND] != 0) {
-	if (strcmp(argv[OPTIND], "-q") == 0) {	/* not getopt compatible */
-	    optind++;
-	    continue;
-	}
-	if ((c = GETOPT(argc, argv, GETOPT_LIST)) <= 0)
-	    break;
-	if (c == 'C') {
-	    VSTRING *buf = vstring_alloc(1);
-
-	    if (setenv(CONF_ENV_PATH,
-		   strcmp(sane_basename(buf, optarg), MAIN_CONF_FILE) == 0 ?
-		       sane_dirname(buf, optarg) : optarg, 1) < 0)
-		msg_fatal_status(EX_UNAVAILABLE, "out of memory");
-	    vstring_free(buf);
-	}
-    }
-    optind = saved_optind;
     mail_conf_read();
     get_mail_conf_str_table(str_table);
 
@@ -1064,7 +1042,7 @@ int     main(int argc, char **argv)
 	    optind++;
 	    continue;
 	}
-	if ((c = GETOPT(argc, argv, GETOPT_LIST)) <= 0)
+	if ((c = GETOPT(argc, argv, "A:B:C:F:GIL:N:O:R:UV:X:b:ce:f:h:imno:p:r:q:tvx")) <= 0)
 	    break;
 	switch (c) {
 	default:
