@@ -44,20 +44,19 @@
 /* .IP type
 /*	The type determines the arguments that follow.
 /* .RS
-/* .IP "ATTR_TYPE_INT (char *, int)"
+/* .IP "ATTR_TYPE_NUM (char *, int)"
 /*	This argument is followed by an attribute name and an integer.
 /* .IP "ATTR_TYPE_LONG (char *, long)"
 /*	This argument is followed by an attribute name and a long integer.
 /* .IP "ATTR_TYPE_STR (char *, char *)"
 /*	This argument is followed by an attribute name and a null-terminated
 /*	string.
-/* .IP "ATTR_TYPE_DATA (char *, ssize_t, char *)"
+/* .IP "ATTR_TYPE_DATA (char *, int, char *)"
 /*	This argument is followed by an attribute name, an attribute value
 /*	length, and an attribute value pointer.
-/* .IP "ATTR_TYPE_FUNC (ATTR_PRINT_SLAVE_FN, void *)"
+/* .IP "ATTR_TYPE_FUNC (ATTR_PRINT_FN, void *)"
 /*	This argument is followed by a function pointer and generic data
-/*	pointer. The caller-specified function returns whatever the
-/*	specified attribute printing function returns.
+/*	pointer.
 /* .IP "ATTR_TYPE_HASH (HTABLE *)"
 /* .IP "ATTR_TYPE_NAMEVAL (NVTABLE *)"
 /*	The content of the table is sent as a sequence of string-valued
@@ -113,9 +112,9 @@ int     attr_vprint0(VSTREAM *fp, int flags, va_list ap)
     char   *str_val;
     HTABLE_INFO **ht_info_list;
     HTABLE_INFO **ht;
-    ssize_t len_val;
+    int     len_val;
     static VSTRING *base64_buf;
-    ATTR_PRINT_SLAVE_FN print_fn;
+    ATTR_PRINT_FN print_fn;
     void   *print_arg;
 
     /*
@@ -130,7 +129,7 @@ int     attr_vprint0(VSTREAM *fp, int flags, va_list ap)
      */
     while ((attr_type = va_arg(ap, int)) != ATTR_TYPE_END) {
 	switch (attr_type) {
-	case ATTR_TYPE_INT:
+	case ATTR_TYPE_NUM:
 	    attr_name = va_arg(ap, char *);
 	    vstream_fwrite(fp, attr_name, strlen(attr_name) + 1);
 	    int_val = va_arg(ap, int);
@@ -159,20 +158,19 @@ int     attr_vprint0(VSTREAM *fp, int flags, va_list ap)
 	case ATTR_TYPE_DATA:
 	    attr_name = va_arg(ap, char *);
 	    vstream_fwrite(fp, attr_name, strlen(attr_name) + 1);
-	    len_val = va_arg(ap, ssize_t);
+	    len_val = va_arg(ap, int);
 	    str_val = va_arg(ap, char *);
 	    if (base64_buf == 0)
 		base64_buf = vstring_alloc(10);
 	    base64_encode(base64_buf, str_val, len_val);
 	    vstream_fwrite(fp, STR(base64_buf), LEN(base64_buf) + 1);
 	    if (msg_verbose)
-		msg_info("send attr %s = [data %ld bytes]",
-			 attr_name, (long) len_val);
+		msg_info("send attr %s = [data %d bytes]", attr_name, len_val);
 	    break;
 	case ATTR_TYPE_FUNC:
-	    print_fn = va_arg(ap, ATTR_PRINT_SLAVE_FN);
+	    print_fn = va_arg(ap, ATTR_PRINT_FN);
 	    print_arg = va_arg(ap, void *);
-	    print_fn(attr_print0, fp, flags | ATTR_FLAG_MORE, print_arg);
+	    print_fn(fp, flags | ATTR_FLAG_MORE, print_arg);
 	    break;
 	case ATTR_TYPE_HASH:
 	    ht_info_list = htable_list(va_arg(ap, HTABLE *));
@@ -222,14 +220,14 @@ int     main(int unused_argc, char **argv)
     htable_enter(table, "foo-name", mystrdup("foo-value"));
     htable_enter(table, "bar-name", mystrdup("bar-value"));
     attr_print0(VSTREAM_OUT, ATTR_FLAG_NONE,
-		ATTR_TYPE_INT, ATTR_NAME_INT, 4711,
+		ATTR_TYPE_NUM, ATTR_NAME_NUM, 4711,
 		ATTR_TYPE_LONG, ATTR_NAME_LONG, 1234,
 		ATTR_TYPE_STR, ATTR_NAME_STR, "whoopee",
 		ATTR_TYPE_DATA, ATTR_NAME_DATA, strlen("whoopee"), "whoopee",
 		ATTR_TYPE_HASH, table,
 		ATTR_TYPE_END);
     attr_print0(VSTREAM_OUT, ATTR_FLAG_NONE,
-		ATTR_TYPE_INT, ATTR_NAME_INT, 4711,
+		ATTR_TYPE_NUM, ATTR_NAME_NUM, 4711,
 		ATTR_TYPE_LONG, ATTR_NAME_LONG, 1234,
 		ATTR_TYPE_STR, ATTR_NAME_STR, "whoopee",
 		ATTR_TYPE_DATA, ATTR_NAME_DATA, strlen("whoopee"), "whoopee",
