@@ -22,15 +22,7 @@
 /*	const char *names;
 /*	int	flags;
 /*
-/*	int	name_mask_delim_opt(context, table, names, delim, flags)
-/*	const char *context;
-/*	NAME_MASK *table;
-/*	const char *names;
-/*	const char *delim;
-/*	int	flags;
-/*
-/*	const char *str_name_mask_opt(buf, context, table, mask, flags)
-/*	VSTRING	*buf;
+/*	const char *str_name_mask_opt(context, table, mask, flags)
 /*	const char *context;
 /*	NAME_MASK *table;
 /*	int	mask;
@@ -46,12 +38,9 @@
 /*	upon each call.
 /*
 /*	name_mask_opt() and str_name_mask_opt() are extended versions
-/*	with additional fine control. name_mask_delim_opt() supports
-/*	non-default delimiter characters.
+/*	with additional fine control.
 /*
 /*	Arguments:
-/* .IP buf
-/*	Null pointer or pointer to buffer storage.
 /* .IP context
 /*	What kind of names and
 /*	masks are being manipulated, in order to make error messages
@@ -65,8 +54,6 @@
 /*	A bit mask.
 /* .IP flags
 /*	Bit-wise OR of zero or more of the following:
-/* .IP delim
-/*	Delimiter characters to use instead of whitespace and commas.
 /* .RS
 /* .IP NAME_MASK_FATAL
 /*	Require that all names listed in \fIname\fR exist in \fItable\fR,
@@ -77,19 +64,14 @@
 /* .IP NAME_MASK_RETURN
 /*	Require that all names listed in \fIname\fR exist in \fItable\fR,
 /*	and that all bits listed in \fImask\fR exist in \fItable\fR.
-/*	Log a warning, and return 0 (name_mask()) or a null pointer
-/*	(str_name_mask()) if this condition is not met.
-/* .IP NAME_MASK_NUMBER
-/*	Require that all bits listed in \fImask\fR exist in \fItable\fR.
-/*	For unrecognized bits, print the numerical hexadecimal form.
+/*	Return 0 (name_mask()) or a null pointer (str_name_mask())
+/*	if this condition is not met.
 /* .IP NAME_MASK_ANY_CASE
 /*	Enable case-insensitive matching.
 /*	This feature is not enabled by default when calling name_mask();
 /*	it has no effect with str_name_mask().
 /* .IP NAME_MASK_COMMA
 /*	Use comma instead of space when converting a mask to string.
-/* .IP NAME_MASK_PIPE
-/*	Use "|" instead of space when converting a mask to string.
 /* .RE
 /*	The value NAME_MASK_NONE explicitly requests no features,
 /*	and NAME_MASK_DEFAULT enables the default options.
@@ -127,12 +109,12 @@
 
 #define STR(x) vstring_str(x)
 
-/* name_mask_delim_opt - compute mask corresponding to list of names */
+/* name_mask_opt - compute mask corresponding to list of names */
 
-int     name_mask_delim_opt(const char *context, NAME_MASK *table,
-		            const char *names, const char *delim, int flags)
+int     name_mask_opt(const char *context, NAME_MASK *table, const char *names,
+		              int flags)
 {
-    const char *myname = "name_mask";
+    char   *myname = "name_mask";
     char   *saved_names = mystrdup(names);
     char   *bp = saved_names;
     int     result = 0;
@@ -149,17 +131,14 @@ int     name_mask_delim_opt(const char *context, NAME_MASK *table,
      * Break up the names string, and look up each component in the table. If
      * the name is found, merge its mask with the result.
      */
-    while ((name = mystrtok(&bp, delim)) != 0) {
+    while ((name = mystrtok(&bp, ", \t\r\n")) != 0) {
 	for (np = table; /* void */ ; np++) {
 	    if (np->name == 0) {
 		if (flags & NAME_MASK_FATAL)
 		    msg_fatal("unknown %s value \"%s\" in \"%s\"",
 			      context, name, names);
-		if (flags & NAME_MASK_RETURN) {
-		    msg_warn("unknown %s value \"%s\" in \"%s\"",
-			     context, name, names);
+		if (flags & NAME_MASK_RETURN)
 		    return (0);
-		}
 		break;
 	    }
 	    if (lookup(name, np->name) == 0) {
@@ -176,36 +155,27 @@ int     name_mask_delim_opt(const char *context, NAME_MASK *table,
 
 /* str_name_mask_opt - mask to string */
 
-const char *str_name_mask_opt(VSTRING *buf, const char *context,
-			              NAME_MASK *table,
+const char *str_name_mask_opt(const char *context, NAME_MASK *table,
 			              int mask, int flags)
 {
-    const char *myname = "name_mask";
+    char   *myname = "name_mask";
     NAME_MASK *np;
     int     len;
-    static VSTRING *my_buf = 0;
-    int     delim = (flags & NAME_MASK_COMMA ? ',' :
-		     (flags & NAME_MASK_PIPE ? '|' : ' '));
+    static VSTRING *buf = 0;
+    int     delim = (flags & NAME_MASK_COMMA ? ',' : ' ');
+
+    if (buf == 0)
+	buf = vstring_alloc(1);
 
-    if (buf == 0) {
-	if (my_buf == 0)
-	    my_buf = vstring_alloc(1);
-	buf = my_buf;
-    }
     VSTRING_RESET(buf);
 
     for (np = table; mask != 0; np++) {
 	if (np->name == 0) {
-	    if (flags & NAME_MASK_FATAL) {
-		msg_fatal("%s: unknown %s bit in mask: 0x%x",
+	    if (flags & NAME_MASK_FATAL)
+		msg_fatal("%s: invalid %s bit in mask: 0x%x",
 			  myname, context, mask);
-	    } else if (flags & NAME_MASK_RETURN) {
-		msg_warn("%s: unknown %s bit in mask: 0x%x",
-			 myname, context, mask);
+	    if (flags & NAME_MASK_RETURN)
 		return (0);
-	    } else if (flags & NAME_MASK_NUMBER) {
-		vstring_sprintf_append(buf, "0x%x%c", mask, delim);
-	    }
 	    break;
 	}
 	if (mask & np->mask) {
@@ -225,7 +195,7 @@ const char *str_name_mask_opt(VSTRING *buf, const char *context,
  /*
   * Stand-alone test program.
   */
-#include <stdlib.h>
+
 #include <vstream.h>
 
 int     main(int argc, char **argv)
@@ -243,8 +213,7 @@ int     main(int argc, char **argv)
     while (--argc && *++argv) {
 	mask = name_mask("test", table, *argv);
 	vstream_printf("%s -> 0x%x -> %s\n",
-		       *argv, mask, str_name_mask((VSTRING *) 0, "mask_test",
-						  table, mask));
+		       *argv, mask, str_name_mask("mask_test", table, mask));
 	vstream_fflush(VSTREAM_OUT);
     }
     vstring_free(buf);
